def global_func():
    return

class some_class():
    def __init__(self):
        bar = None
        self.foo = bar

    def member_func1(self):
        baz = None
        return baz

    def member_func2(self):
        baz = None
        return baz